@@ -1,5 +1,6 @@
 /* Extended cpio header from POSIX.1.
-   Copyright (C) 1992, 2006, 2007, 2010 Free Software Foundation, Inc.
+   Copyright (C) 1992, 2006-2007, 2010, 2014-2015 Free Software
+   Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -123,7 +124,7 @@ struct cpio_file_stat /* Internal representation of a CPIO header */
   long c_rdev_maj;
   long c_rdev_min;
   size_t c_namesize;
-  unsigned long c_chksum;
+  uint32_t c_chksum;
   char *c_name;
   char *c_tar_linkname;
 };
